(**
  To be viewed in Isabelle/jedit IDE. In a normal text editor, this will just be UNREADABLE!

  http://isabelle.in.tum.de

*)
theory Edka_Abstract_Demo
  (*
    We import some background theory
  *)
  imports 
    EdmondsKarp_Maxflow.FordFulkerson_Algo 
    EdmondsKarp_Maxflow.EdmondsKarp_Termination_Abstract
begin

(* Shut up some confusing warnings \<dots> *)
no_notation Heap_Monad.return ("return")


(*
  We fix a flow network:
*)
context Network begin

definition edmonds_karp where "edmonds_karp \<equiv> do {
  let f = (\<lambda>_. 0);
  (f,_) \<leftarrow> while\<^sub>T
    (\<lambda>(f,brk). \<not>brk) 
    (\<lambda>(f,_). do {
      p \<leftarrow> select p. Graph.isShortestPath (residualGraph c f) s p t;
      case p of 
        None \<Rightarrow> return (f,True)
      | Some p \<Rightarrow> do {
          let f = NFlow.augment_with_path c f p;
          return (f, False)
        }  
    })
    (f,False);
  return f 
}"

text \<open>From the background theory, we get a measure function on the residual graph, 
  that decreases when augmenting with a shortest path\<close>
term ek_analysis_defs.ekMeasure
thm NFlow.shortest_path_decr_ek_measure

text \<open>The invariant states that
  \<^item> We always have a valid flow
  \<^item> If we break, there is no augmenting path
\<close>
abbreviation edka_invar where "edka_invar \<equiv> \<lambda>(f,brk). 
        NFlow c s t f 
      \<and> (brk \<longrightarrow> (\<forall>p. \<not>NPreflow.isAugmentingPath c s t f p))
    "  

text \<open>Based on the measure function, it is easy to obtain a well-founded 
  relation that proves termination of the loop in the Edmonds-Karp algorithm:\<close>
definition "edka_wf_rel \<equiv> inv_image 
  (less_than_bool <*lex*> measure (\<lambda>cf. ek_analysis_defs.ekMeasure cf s t))
  (\<lambda>(f,brk). (\<not>brk,residualGraph c f))"

lemma edka_wf_rel_wf[simp, intro!]: "wf edka_wf_rel"
  unfolding edka_wf_rel_def by auto

(* Now we can prove the algorithm correct.
  We do the proof by hand first, to illustrate the operation of the VCG
*)
lemma "edmonds_karp \<le> (spec f. isMaxFlow f)"
  unfolding edmonds_karp_def
  apply (rule refine_vcg) (* let *)
  apply (rule refine_vcg) (* bind *)
  apply (rule WHILET_rule[where I=edka_invar and R=edka_wf_rel]) (* For while, we have to specify invariant and variant *)
  apply simp (* relation is well-founded. *)
  apply (simp add: zero_flow) (* Invariant holds initially *)
  apply (rule refine_vcg) (* prod-case *)
  apply (rule refine_vcg) (* bind *)
  apply (rule refine_vcg) (* select *) thm SELECT_rule
  apply (rule refine_vcg) (* option-case *)
  apply simp (* trivial, False \<Longrightarrow> *)
  (* We found a path *)
  apply (rule refine_vcg) (* let *)
  apply (rule refine_vcg) (* return *)
  subgoal (* Invar preserved by step, measure decreases *)
    apply (clarsimp; intro conjI)
    apply (simp add: NFlow.augment_pres_nflow NFlow.augment_with_path_def NFlow.shortest_is_augmenting)
    apply (simp add: edka_wf_rel_def NFlow.shortest_path_decr_ek_measure NFlow.augment_with_path_def)
    done
  apply (rule refine_vcg) (* option-case *)
  (* We did not find a path *)
  apply (rule refine_vcg) (* return *)
  subgoal (* Invariant preserved, measure decreases *)
    apply simp
    apply (simp add: edka_wf_rel_def)
    apply (auto dest: NFlow.augmenting_path_imp_shortest) (* If there is no shortest augmenting path, there cannot be an augmenting path*)
    done
  apply simp (* trivial, False \<Longrightarrow> *)

  apply (rule refine_vcg) (* prod-case *)
  apply (rule refine_vcg) (* return *)
  subgoal (* Invariant and \<not>loop-condition implies goal *)
    apply clarsimp
    thm NFlow.noAugPath_iff_maxFlow[symmetric]
    apply (simp add: NFlow.noAugPath_iff_maxFlow[symmetric])
    done
  done


(* The VCG automatically does all the straightforward steps for us. 
  We only have to specify the invariant and variant for the loop. *)
lemma "edmonds_karp \<le> (spec f. isMaxFlow f)"
  unfolding edmonds_karp_def
  apply (refine_vcg WHILET_rule[where I=fofu_invar and R=edka_wf_rel])
  apply clarsimp_all (* Get rid of the trivial subgoals *)
  (* What remains are the goals that involve some background theory. 
    In particular, they do not contain any nres monad stuff. *)

  (*
    There are many ways to find the proofs of these high-level VCs.
    The following has been found (almost) automatically by sledgehammer.

    Sledgehammer translates the goal and the background theory to a problem for SMT solvers and for FOL-solvers,
    invokes various solvers, and then translates back the found proofs to Isabelle/HOL. 
    You do NOT need to trust the SMT-solvers!

    In practice, you will start to extend the background theory at this point!
  *)

  subgoal by (simp add: zero_flow)
  subgoal by (simp add: NFlow.augment_pres_nflow NFlow.augment_with_path_def NFlow.shortest_is_augmenting)
  subgoal unfolding edka_wf_rel_def apply simp
    by (simp add: NFlow.augment_with_path_def NFlow.shortest_path_decr_ek_measure)
  subgoal using NFlow.augmenting_path_imp_shortest by blast
  subgoal unfolding edka_wf_rel_def by simp
  subgoal using NFlow.noAugPath_iff_maxFlow by blast

  (*
    Alternative proof, knowing the required lemmas from the background theory:

  apply (auto simp: 
    zero_flow 
    NFlow.augment_pres_nflow 
    NFlow.augmenting_path_not_empty
    NFlow.noAugPath_iff_maxFlow[symmetric]
    NFlow.augment_with_path_def
    edka_wf_rel_def NFlow.shortest_path_decr_ek_measure
    NFlow.shortest_is_augmenting
    dest: NFlow.augmenting_path_imp_shortest
  )
  *)
  done

end  

end
